// wow.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#define WIN32_NO_STATUS   /* Tell Windows headers you'll use ntstatus.s from NDK */
#include "windows.h"      /* Declare Windows Headers like you normally would */
#include "ntndk.h"        /* Declare the NDK Headers */
#include "wow.h"
#include "Ntstrsafe.h"

 

/* GLOBALS ********************************************************************/

UNICODE_STRING SmpSubsystemName, PosixName, Os2Name;
LIST_ENTRY SmpBootExecuteList, SmpSetupExecuteList, SmpPagingFileList;
LIST_ENTRY SmpDosDevicesList, SmpFileRenameList, SmpKnownDllsList;
LIST_ENTRY SmpExcludeKnownDllsList, SmpSubSystemList, SmpSubSystemsToLoad;
LIST_ENTRY SmpSubSystemsToDefer, SmpExecuteList, NativeProcessList;

PVOID SmpHeap;
ULONG SmBaseTag = HEAP_ZERO_MEMORY;
HANDLE SmpDebugPort, SmpDosDevicesObjectDirectory;
PWCHAR SmpDefaultEnvironment, SmpDefaultLibPathBuffer;
UNICODE_STRING SmpKnownDllPath, SmpDefaultLibPath;
ULONG SmpCalledConfigEnv;

ULONG SmpInitProgressByLine;
NTSTATUS SmpInitReturnStatus;
PVOID SmpInitLastCall;

SECURITY_DESCRIPTOR SmpPrimarySDBody, SmpLiberalSDBody, SmpKnownDllsSDBody;
SECURITY_DESCRIPTOR SmpApiPortSDBody;
PISECURITY_DESCRIPTOR SmpPrimarySecurityDescriptor, SmpLiberalSecurityDescriptor;
PISECURITY_DESCRIPTOR SmpKnownDllsSecurityDescriptor, SmpApiPortSecurityDescriptor;

ULONG SmpAllowProtectedRenames, SmpProtectionMode = 1;
BOOLEAN MiniNTBoot;
WCHAR pszDest[50];
size_t cchDest = 50;

#define SMSS_CHECKPOINT(x, y)           \
{                                       \
    SmpInitProgressByLine = __LINE__;   \
    SmpInitReturnStatus = (y);          \
    SmpInitLastCall = (x);              \
}


NTSTATUS
NTAPI
SmpSaveRegistryValue(IN PLIST_ENTRY ListAddress,
	IN PWSTR Name,
	IN PWCHAR Value,
	IN BOOLEAN Flags)
{
	PSMP_REGISTRY_VALUE RegEntry;
	UNICODE_STRING NameString, ValueString;
	ANSI_STRING AnsiValueString;
	PLIST_ENTRY NextEntry;

	/* Convert to unicode strings */
	RtlInitUnicodeString(&NameString, Name);
	RtlInitUnicodeString(&ValueString, Value);

	/* In case this is the first value, initialize a new list/structure */
	RegEntry = NULL;

	/* Check if we should do a duplicate check */
	if (Flags)
	{
		/* Loop the current list */
		NextEntry = ListAddress->Flink;
		while (NextEntry != ListAddress)
		{
			/* Get each entry */
			RegEntry = CONTAINING_RECORD(NextEntry, SMP_REGISTRY_VALUE, Entry);

			/* Check if the value name matches */
			if (!RtlCompareUnicodeString(&RegEntry->Name, &NameString, TRUE))
			{
				/* Check if the value is the exact same thing */
				if (!RtlCompareUnicodeString(&RegEntry->Value, &ValueString, TRUE))
				{
					/* Fail -- the same setting is being set twice */
					return STATUS_OBJECT_NAME_EXISTS;
				}

				/* We found the list, and this isn't a duplicate value */
				break;
			}

			/* This wasn't a match, keep going */
			NextEntry = NextEntry->Flink;
			RegEntry = NULL;
		}
	}

	/* Are we adding on, or creating a new entry */
	if (!RegEntry)
	{
		/* A new entry -- allocate it */
		RegEntry = (PSMP_REGISTRY_VALUE)RtlAllocateHeap(RtlGetProcessHeap(),
			SmBaseTag,
			sizeof(SMP_REGISTRY_VALUE) +
			NameString.MaximumLength);
		if (!RegEntry) return STATUS_NO_MEMORY;

		/* Initialize the list and set all values to NULL */
		InitializeListHead(&RegEntry->Entry);
		RegEntry->AnsiValue = NULL;
		RegEntry->Value.Buffer = NULL;

		/* Copy and initialize the value name */
		RegEntry->Name.Buffer = (PWCHAR)(RegEntry + 1);
		RegEntry->Name.Length = NameString.Length;
		RegEntry->Name.MaximumLength = NameString.MaximumLength;
		RtlCopyMemory(RegEntry->Name.Buffer,
			NameString.Buffer,
			NameString.MaximumLength);

		/* Add this entry into the list */
		InsertTailList(ListAddress, &RegEntry->Entry);
	}

	/* Did we have an old value buffer? */
	if (RegEntry->Value.Buffer)
	{
		/* Free it */
		ASSERT(RegEntry->Value.Length != 0);
		RtlFreeHeap(RtlGetProcessHeap(), 0, RegEntry->Value.Buffer);
	}

	/* Is there no value associated? */
	if (!Value)
	{
		/* We're done here */
		RtlInitUnicodeString(&RegEntry->Value, NULL);
		return STATUS_SUCCESS;
	}

	/* There is a value, so allocate a buffer for it */
	RegEntry->Value.Buffer = (PWSTR)RtlAllocateHeap(RtlGetProcessHeap(),
		SmBaseTag,
		ValueString.MaximumLength);
	if (!RegEntry->Value.Buffer)
	{
		/* Out of memory, undo */
		RemoveEntryList(&RegEntry->Entry);
		RtlFreeHeap(RtlGetProcessHeap(), 0, RegEntry);
		return STATUS_NO_MEMORY;
	}

	/* Copy the value into the entry */
	RegEntry->Value.Length = ValueString.Length;
	RegEntry->Value.MaximumLength = ValueString.MaximumLength;
	RtlCopyMemory(RegEntry->Value.Buffer,
		ValueString.Buffer,
		ValueString.MaximumLength);

	/* Now allocate memory for an ANSI copy of it */
	RegEntry->AnsiValue = (PCHAR)RtlAllocateHeap(RtlGetProcessHeap(),
		SmBaseTag,
		(ValueString.Length / sizeof(WCHAR)) +
		sizeof(ANSI_NULL));
	if (!RegEntry->AnsiValue)
	{
		/* Out of memory, undo */
		RtlFreeHeap(RtlGetProcessHeap(), 0, RegEntry->Value.Buffer);
		RemoveEntryList(&RegEntry->Entry);
		RtlFreeHeap(RtlGetProcessHeap(), 0, RegEntry);
		return STATUS_NO_MEMORY;
	}

	/* Convert the Unicode value string and return success */
	RtlInitEmptyAnsiString(&AnsiValueString,
		RegEntry->AnsiValue,
		(ValueString.Length / sizeof(WCHAR)) +
		sizeof(ANSI_NULL));
	RtlUnicodeStringToAnsiString(&AnsiValueString, &ValueString, FALSE);
	return STATUS_SUCCESS;
}

PSMP_REGISTRY_VALUE
NTAPI
SmpFindRegistryValue(IN PLIST_ENTRY List,
	IN PWSTR ValueName)
{
	PSMP_REGISTRY_VALUE RegEntry = NULL;
	UNICODE_STRING ValueString;
	PLIST_ENTRY NextEntry;

	/* Initialize the value name sting */
	RtlInitUnicodeString(&ValueString, ValueName);

	/* Loop the list */
	NextEntry = List->Flink;
	while (NextEntry != List)
	{
		/* Get each entry */
		RegEntry = CONTAINING_RECORD(NextEntry, SMP_REGISTRY_VALUE, Entry);

		/* Check if the value name matches */
		if (!RtlCompareUnicodeString(&RegEntry->Name, &ValueString, TRUE)) break;

		/* It doesn't, move on */
		NextEntry = NextEntry->Flink;
	}

	/* If we looped back, return NULL, otherwise return the entry we found */
	if (NextEntry == List) RegEntry = NULL;
	return RegEntry;
}

NTSTATUS
NTAPI
SmpInitializeKnownDllPath(IN PUNICODE_STRING DllPath,
	IN PWCHAR Buffer,
	IN ULONG Length)
{
	NTSTATUS Status;

	/* Allocate the buffer */
	DllPath->Buffer = (PWSTR)RtlAllocateHeap(RtlGetProcessHeap(), SmBaseTag, Length);
	if (DllPath->Buffer)
	{
		/* Fill out the rest of the string */
		DllPath->MaximumLength = (USHORT)Length;
		DllPath->Length = (USHORT)Length - sizeof(UNICODE_NULL);

		/* Copy the actual path and return success */
		RtlCopyMemory(DllPath->Buffer, Buffer, Length);
		Status = STATUS_SUCCESS;
	}
	else
	{
		/* Fail with out of memory code */
		Status = STATUS_NO_MEMORY;
	}

	/* Return result */
	return Status;
}

NTSTATUS
NTAPI
SmpConfigureKnownDlls(IN PWSTR ValueName,
	IN ULONG ValueType,
	IN PVOID ValueData,
	IN ULONG ValueLength,
	IN PVOID Context,
	IN PVOID EntryContext)
{
	/* Check which value is being set */
	if (_wcsicmp(ValueName, L"DllDirectory32") == 0|| _wcsicmp(ValueName, L"DllDirectory") == 0)
	{
		
		/* This is the directory, initialize it */
		return RtlStringCchPrintfW(pszDest, cchDest, L"The %ls Path is: %ls\n",ValueName,(PWCHAR)ValueData);
		 //SmpInitializeKnownDllPath(&SmpKnownDllPath, (PWCHAR)ValueData, ValueLength);
	}
	else
	{
		/* Add to the linked list -- this is a file */
		return SmpSaveRegistryValue((PLIST_ENTRY)EntryContext, ValueName, (PWCHAR)ValueData, TRUE);
	}
}

NTSTATUS
NTAPI
SmpConfigureExcludeKnownDlls(IN PWSTR ValueName,
	IN ULONG ValueType,
	IN PVOID ValueData,
	IN ULONG ValueLength,
	IN PVOID Context,
	IN PVOID EntryContext)
{
	PWCHAR DllName;
	NTSTATUS Status;

	/* Make sure the value type is valid */
	if ((ValueType == REG_MULTI_SZ) || (ValueType == REG_SZ))
	{
		/* Keep going for each DLL in the list */
		DllName = (PWCHAR)ValueData;
		while (*DllName)
		{
			/* Add this to the linked list */
			//wprintf(L"Excluded DLL: %lS\n", DllName);
			Status = SmpSaveRegistryValue((PLIST_ENTRY)EntryContext, DllName, NULL, TRUE);

			/* Bail out on failure or if only one DLL name was present */
			if (!(NT_SUCCESS(Status)) || (ValueType == REG_SZ)) return Status;

			/* Otherwise, move to the next DLL name */
			while (*DllName++);
		}
	}

	/* All done */
	return STATUS_SUCCESS;
}

RTL_QUERY_REGISTRY_TABLE
SmpRegistryConfigurationTable[] =
{
	
	{
		SmpConfigureExcludeKnownDlls,
		0,
	L"ExcludeFromKnownDlls",
	&SmpExcludeKnownDllsList,
	REG_MULTI_SZ,
	L"\0",
	0
	},
		
	{
		SmpConfigureKnownDlls,
		RTL_QUERY_REGISTRY_SUBKEY,
	L"KnownDlls",
	&SmpKnownDllsList,
	REG_NONE,
	NULL,
	0
	},

	{ 0 },
};

/* REGISTRY CONFIGURATION *****************************************************/
NTSTATUS
NTAPI
SmpTerminate(IN PULONG_PTR Parameters,
	IN ULONG ParameterMask,
	IN ULONG ParameterCount)
{
	NTSTATUS Status;
	BOOLEAN Old;
	ULONG Response;

	/* Give the shutdown privilege to the thread */
	if (RtlAdjustPrivilege(SE_SHUTDOWN_PRIVILEGE, TRUE, TRUE, &Old) ==
		STATUS_NO_TOKEN)
	{
		/* Thread doesn't have a token, give it to the entire process */
		RtlAdjustPrivilege(SE_SHUTDOWN_PRIVILEGE, TRUE, FALSE, &Old);
	}

	/* Take down the process/machine with a hard error */
	Status = NtRaiseHardError(STATUS_SYSTEM_PROCESS_TERMINATED,
		ParameterCount,
		ParameterMask,
		Parameters,
		OptionShutdownSystem,
		&Response);

	/* Terminate the process if the hard error didn't already */
	return NtTerminateProcess(NtCurrentProcess(), Status);
}

VOID
NTAPI
SmpProcessModuleImports(IN PVOID Unused,
	IN PCHAR ImportName)
{
	ULONG Length = 0, Chars;
	WCHAR Buffer[MAX_PATH];
	PWCHAR DllName, DllValue;
	ANSI_STRING ImportString;
	UNICODE_STRING ImportUnicodeString;
	NTSTATUS Status;

	/* Skip NTDLL since it's already always mapped */
	if (!_stricmp(ImportName, "ntdll.dll")) return;

	/* Initialize our strings */
	RtlInitAnsiString(&ImportString, ImportName);
	RtlInitEmptyUnicodeString(&ImportUnicodeString, Buffer, sizeof(Buffer));
	Status = RtlAnsiStringToUnicodeString(&ImportUnicodeString, &ImportString, FALSE);
	if (!NT_SUCCESS(Status)) return;

	/* Loop in case we find a forwarder */
	ImportUnicodeString.MaximumLength = ImportUnicodeString.Length + sizeof(UNICODE_NULL);
	while (Length < ImportUnicodeString.Length)
	{
		if (ImportUnicodeString.Buffer[Length / sizeof(WCHAR)] == L'.') break;
		Length += sizeof(WCHAR);
	}

	/* Break up the values as needed */
	DllValue = ImportUnicodeString.Buffer;
	DllName = &ImportUnicodeString.Buffer[ImportUnicodeString.MaximumLength / sizeof(WCHAR)];
	Chars = Length >> 1;
	RtlStringCchCopyNW(DllName, Chars,ImportUnicodeString.Buffer, Chars);
	DllName[Chars] = 0;

	/* Add the DLL to the list */
	SmpSaveRegistryValue(&SmpKnownDllsList, DllName, DllValue, TRUE);
}

NTSTATUS
NTAPI
SmpCreateSecurityDescriptors(IN BOOLEAN InitialCall)
{
	NTSTATUS Status;
	PSID WorldSid = NULL, AdminSid = NULL, SystemSid = NULL, pAllAppsSID = NULL;
	PSID RestrictedSid = NULL, OwnerSid = NULL;
	SID_IDENTIFIER_AUTHORITY WorldAuthority = { SECURITY_WORLD_SID_AUTHORITY };
	SID_IDENTIFIER_AUTHORITY NtAuthority = { SECURITY_NT_AUTHORITY };
	SID_IDENTIFIER_AUTHORITY CreatorAuthority = { SECURITY_CREATOR_SID_AUTHORITY };
	SID_IDENTIFIER_AUTHORITY ApplicationAuthority = SECURITY_APP_PACKAGE_AUTHORITY;

	ULONG AclLength, SidLength;
	PACL Acl;
	PACE_HEADER Ace = NULL;
	BOOLEAN ProtectionRequired = FALSE;

	/* Check if this is the first call */
	if (InitialCall)
	{
		/* Create and set the primary descriptor */
		SmpPrimarySecurityDescriptor = &SmpPrimarySDBody;
		Status = RtlCreateSecurityDescriptor(SmpPrimarySecurityDescriptor,
			SECURITY_DESCRIPTOR_REVISION);
		ASSERT(NT_SUCCESS(Status));
		Status = RtlSetDaclSecurityDescriptor(SmpPrimarySecurityDescriptor,
			TRUE,
			NULL,
			FALSE);
		ASSERT(NT_SUCCESS(Status));

		/* Create and set the liberal descriptor */
		SmpLiberalSecurityDescriptor = &SmpLiberalSDBody;
		Status = RtlCreateSecurityDescriptor(SmpLiberalSecurityDescriptor,
			SECURITY_DESCRIPTOR_REVISION);
		ASSERT(NT_SUCCESS(Status));
		Status = RtlSetDaclSecurityDescriptor(SmpLiberalSecurityDescriptor,
			TRUE,
			NULL,
			FALSE);
		ASSERT(NT_SUCCESS(Status));

		/* Create and set the \KnownDlls descriptor */
		SmpKnownDllsSecurityDescriptor = &SmpKnownDllsSDBody;
		Status = RtlCreateSecurityDescriptor(SmpKnownDllsSecurityDescriptor,
			SECURITY_DESCRIPTOR_REVISION);
		ASSERT(NT_SUCCESS(Status));
		Status = RtlSetDaclSecurityDescriptor(SmpKnownDllsSecurityDescriptor,
			TRUE,
			NULL,
			FALSE);
		ASSERT(NT_SUCCESS(Status));
	}

	/* Check if protection was requested in the registry (on by default) */
	if (SmpProtectionMode & 1) ProtectionRequired = TRUE;

	/* Exit if there's nothing to do */
	if (!(InitialCall || ProtectionRequired)) return STATUS_SUCCESS;	
	/* Build the world SID */
	Status = RtlAllocateAndInitializeSid(&WorldAuthority, 1,
		SECURITY_WORLD_RID,
		0, 0, 0, 0, 0, 0, 0,
		&WorldSid);
	if (!NT_SUCCESS(Status))
	{
		WorldSid = NULL;
		goto Quickie;
	}

	/* Build the admin SID */
	Status = RtlAllocateAndInitializeSid(&NtAuthority, 2,
		SECURITY_BUILTIN_DOMAIN_RID,
		DOMAIN_ALIAS_RID_ADMINS,
		0, 0, 0, 0, 0, 0,
		&AdminSid);
	if (!NT_SUCCESS(Status))
	{
		AdminSid = NULL;
		goto Quickie;
	}

	/* Build the owner SID */
	Status = RtlAllocateAndInitializeSid(&CreatorAuthority, 1,
		SECURITY_CREATOR_OWNER_RID,
		0, 0, 0, 0, 0, 0, 0,
		&OwnerSid);
	if (!NT_SUCCESS(Status))
	{
		OwnerSid = NULL;
		goto Quickie;
	}

	/* Build the restricted SID */
	Status = RtlAllocateAndInitializeSid(&NtAuthority, 1,
		SECURITY_RESTRICTED_CODE_RID,
		0, 0, 0, 0, 0, 0, 0,
		&RestrictedSid);
	if (!NT_SUCCESS(Status))
	{
		RestrictedSid = NULL;
		goto Quickie;
	}

	/* Build the system SID */
	Status = RtlAllocateAndInitializeSid(&NtAuthority, 1,
		SECURITY_LOCAL_SYSTEM_RID,
		0, 0, 0, 0, 0, 0, 0,
		&SystemSid);
	if (!NT_SUCCESS(Status))
	{
		SystemSid = NULL;
		goto Quickie;
	}

	/* Build the AppX SID */
	Status = RtlAllocateAndInitializeSid(&ApplicationAuthority,
		SECURITY_BUILTIN_APP_PACKAGE_RID_COUNT,
		SECURITY_APP_PACKAGE_BASE_RID,
		SECURITY_BUILTIN_PACKAGE_ANY_PACKAGE,
		0, 0, 0, 0, 0, 0,
		&pAllAppsSID);

	if (!NT_SUCCESS(Status))
	{
		pAllAppsSID = NULL;
		//goto Quickie;
	}
	
	/* Now check if we're creating the core descriptors */
	if (!InitialCall)
	{
		/* We're skipping NextAcl so we have to do this here */
		SidLength = RtlLengthSid(WorldSid) + RtlLengthSid(RestrictedSid) + RtlLengthSid(AdminSid);
		SidLength *= 2;
		goto NotInitial;
	}
	
NextAcl:
	/* Allocate an ACL with 6 ACEs, two ACEs per SID */
	SidLength = RtlLengthSid(WorldSid) + RtlLengthSid(RestrictedSid) + RtlLengthSid(AdminSid) + RtlLengthSid(pAllAppsSID);
	SidLength = 2 * SidLength;
	AclLength = sizeof(ACL) + (8 * sizeof(ACCESS_ALLOWED_ACE)) + SidLength;

	Acl = (PACL)RtlAllocateHeap(RtlGetProcessHeap(), HEAP_ZERO_MEMORY, AclLength);
	if (!Acl) Status = STATUS_NO_MEMORY;
	if (!NT_SUCCESS(Status)) goto NotInitial;

	/* Now build the ACL and add the six ACEs */
	Status = RtlCreateAcl(Acl, AclLength, ACL_REVISION);
	ASSERT(NT_SUCCESS(Status));

	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE, WorldSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE, RestrictedSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_ALL, AdminSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ | GENERIC_WRITE, WorldSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ | GENERIC_WRITE, RestrictedSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_ALL, AdminSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ | GENERIC_WRITE, pAllAppsSID);
	ASSERT(NT_SUCCESS(Status));

	/* Now edit the last three ACEs and make them inheritable */
	Status = RtlGetAce(Acl, 3, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;
	Status = RtlGetAce(Acl, 4, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;
	Status = RtlGetAce(Acl, 5, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;
	Status = RtlGetAce(Acl, 6, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;

	/* Set this as the DACL */
	Status = RtlSetDaclSecurityDescriptor(SmpKnownDllsSecurityDescriptor,
		TRUE,
		Acl,
		FALSE);
	ASSERT(NT_SUCCESS(Status));

NotInitial:
	/* The initial ACLs have been created, are we also protecting objects? */
	if (!ProtectionRequired) goto Quickie;

	/* Allocate an ACL with 7 ACEs, two ACEs per SID, and one final owner ACE */
	SidLength += RtlLengthSid(OwnerSid);
	AclLength = sizeof(ACL) + 7 * sizeof(ACCESS_ALLOWED_ACE) + 2 * SidLength;

	Acl = (PACL)RtlAllocateHeap(RtlGetProcessHeap(), HEAP_ZERO_MEMORY, AclLength);
	if (!Acl) Status = STATUS_NO_MEMORY;
	if (!NT_SUCCESS(Status)) goto Quickie;

	/* Build the ACL and add the seven ACEs */
	Status = RtlCreateAcl(Acl, AclLength, ACL_REVISION);
	ASSERT(NT_SUCCESS(Status));

	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ, WorldSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ, RestrictedSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_ALL, AdminSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ, WorldSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ, RestrictedSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_ALL, AdminSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_ALL, OwnerSid);
	ASSERT(NT_SUCCESS(Status));

	/* Edit the last 4 ACEs to make then inheritable */
	Status = RtlGetAce(Acl, 3, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;
	Status = RtlGetAce(Acl, 4, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;
	Status = RtlGetAce(Acl, 5, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;
	Status = RtlGetAce(Acl, 6, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;

	/* Set this as the DACL for the primary SD */
	Status = RtlSetDaclSecurityDescriptor(SmpPrimarySecurityDescriptor,
		TRUE,
		Acl,
		FALSE);
	ASSERT(NT_SUCCESS(Status));

	/* Allocate an ACL with 7 ACEs, two ACEs per SID, and one final owner ACE */
	AclLength = sizeof(ACL) + 7 * sizeof(ACCESS_ALLOWED_ACE) + 2 * SidLength;
	Acl = (PACL)RtlAllocateHeap(RtlGetProcessHeap(), HEAP_ZERO_MEMORY, AclLength);
	if (!Acl) Status = STATUS_NO_MEMORY;
	if (!NT_SUCCESS(Status)) goto Quickie;

	/* Build the ACL and add the seven ACEs */
	Status = RtlCreateAcl(Acl, AclLength, ACL_REVISION);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ | GENERIC_WRITE, WorldSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ | GENERIC_WRITE, RestrictedSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_ALL, AdminSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ | GENERIC_WRITE, WorldSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_EXECUTE | GENERIC_READ | GENERIC_WRITE, RestrictedSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_ALL, AdminSid);
	ASSERT(NT_SUCCESS(Status));
	Status = RtlAddAccessAllowedAce(Acl, ACL_REVISION, GENERIC_ALL, OwnerSid);
	ASSERT(NT_SUCCESS(Status));

	/* Edit the last 4 ACEs to make then inheritable */
	Status = RtlGetAce(Acl, 3, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;
	Status = RtlGetAce(Acl, 4, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;
	Status = RtlGetAce(Acl, 5, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;
	Status = RtlGetAce(Acl, 6, (PVOID *)&Ace);
	ASSERT(NT_SUCCESS(Status));
	Ace->AceFlags = OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE | INHERIT_ONLY_ACE;

	/* Now set this as the DACL for the liberal SD */
	Status = RtlSetDaclSecurityDescriptor(SmpLiberalSecurityDescriptor,
		TRUE,
		Acl,
		FALSE);
	ASSERT(NT_SUCCESS(Status));

Quickie:
	/* Cleanup the SIDs */
	if (OwnerSid) RtlFreeHeap(RtlGetProcessHeap(), 0, OwnerSid);
	if (AdminSid) RtlFreeHeap(RtlGetProcessHeap(), 0, AdminSid);
	if (WorldSid) RtlFreeHeap(RtlGetProcessHeap(), 0, WorldSid);
	if (SystemSid) RtlFreeHeap(RtlGetProcessHeap(), 0, SystemSid);
	if (RestrictedSid) RtlFreeHeap(RtlGetProcessHeap(), 0, RestrictedSid);
	
	if (pAllAppsSID) RtlFreeHeap(RtlGetProcessHeap(), 0, pAllAppsSID);

	return Status;
}

NTSTATUS
NTAPI
SmpInitializeKnownDllsInternal(IN PUNICODE_STRING Directory,
	IN PUNICODE_STRING Path)
{
	HANDLE DirFileHandle, DirHandle, SectionHandle, FileHandle, LinkHandle;
	UNICODE_STRING NtPath, DestinationString;
	OBJECT_ATTRIBUTES ObjectAttributes;
	NTSTATUS Status, Status1;
	PLIST_ENTRY NextEntry;
	PSMP_REGISTRY_VALUE RegEntry;
	ULONG_PTR ErrorParameters[3];
	UNICODE_STRING ErrorResponse;
	IO_STATUS_BLOCK IoStatusBlock;
	SECURITY_DESCRIPTOR_CONTROL OldFlag = 0;
	USHORT ImageCharacteristics;
	PACL pAcl = NULL;
	//PSE_EXPORTS SeExports;
	ULONG aclSize = 0;
	SECURITY_DESCRIPTOR sd = { 0 };

	//TFNNtOpenFile pfnNtOpenFile = NULL;
	//TFNLdrVerifyImageMatchesChecksum pfnLdrVerifyImageMatchesChecksum = NULL;
	//TFNNtCreateSection pfnNtCreateSection = NULL;

	/* Initialize to NULL */
	DirFileHandle = NULL;
	DirHandle = NULL;
	NtPath.Buffer = NULL;
	
	/* Create the \KnownDLLs32 directory */
	InitializeObjectAttributes(&ObjectAttributes,
		Directory,
		OBJ_CASE_INSENSITIVE | OBJ_OPENIF | OBJ_PERMANENT,
		NULL,
		SmpKnownDllsSecurityDescriptor);
	Status = NtCreateDirectoryObject(&DirHandle,
		DIRECTORY_ALL_ACCESS,
		&ObjectAttributes);
	if (!NT_SUCCESS(Status))
	{
		/* Handle failure */
		RtlStringCchPrintfW(pszDest, cchDest, L"SMSS: Unable to create %ls directory - Status == %lx\n",
			Directory->Buffer, Status);
		return Status;
	}
	/* Convert the path to native format */
	if (!RtlDosPathNameToNtPathName_U(Path->Buffer, &NtPath, NULL, NULL))
	{
		/* Fail if this didn't work */
		RtlStringCchPrintfW(pszDest, cchDest, L"SMSS: Unable to to convert %ls to an Nt path\n", Path->Buffer);
		Status = STATUS_OBJECT_NAME_INVALID;
		goto Quickie;
	}

	/* Open the path that was specified, which should be a directory */
	InitializeObjectAttributes(&ObjectAttributes,
		&NtPath,
		OBJ_CASE_INSENSITIVE,
		NULL,
		NULL);
	Status = NtOpenFile(&DirFileHandle,
		FILE_LIST_DIRECTORY | SYNCHRONIZE,
		&ObjectAttributes,
		&IoStatusBlock,
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		FILE_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_NONALERT);
	if (!NT_SUCCESS(Status))
	{
		/* Fail if we couldn't open it */
		RtlStringCchPrintfW(pszDest, cchDest, L"SMSS: Unable to open a handle to the KnownDll directory (%ls)"
			L"- Status == %lx\n",
			Path->Buffer,
			Status);
		FileHandle = NULL;
		goto Quickie;
	}

	/* Temporarily hack the SD to use a default DACL for this symbolic link */
	if (SmpPrimarySecurityDescriptor)
	{
		OldFlag = SmpPrimarySecurityDescriptor->Control;
		SmpPrimarySecurityDescriptor->Control |= SE_DACL_DEFAULTED;
	}

	/* Create a symbolic link to the directory in the object manager */
	RtlInitUnicodeString(&DestinationString, L"KnownDllPath");
	InitializeObjectAttributes(&ObjectAttributes,
		&DestinationString,
		OBJ_CASE_INSENSITIVE | OBJ_OPENIF | OBJ_PERMANENT,
		DirHandle,
		SmpPrimarySecurityDescriptor);
	Status = NtCreateSymbolicLinkObject(&LinkHandle,
		SYMBOLIC_LINK_ALL_ACCESS,
		&ObjectAttributes,
		Path);

	/* Undo the hack */
	if (SmpPrimarySecurityDescriptor) SmpPrimarySecurityDescriptor->Control = OldFlag;

	/* Check if the symlink was created */
	if (!NT_SUCCESS(Status))
	{
		/* It wasn't, so bail out since the OS needs it to exist */
		RtlStringCchPrintfW(pszDest, cchDest, L"SMSS: Unable to create %ls symbolic link - Status == %lx\n",
			DestinationString.Buffer , Status);
		LinkHandle = NULL;
		goto Quickie;
	}

	/* We created it permanent, we can go ahead and close the handle now */
	Status1 = NtClose(LinkHandle);
	ASSERT(NT_SUCCESS(Status1));
	
	/* Now loop the known DLLs */
	NextEntry = SmpKnownDllsList.Flink;
	while (NextEntry != &SmpKnownDllsList)
	{
		/* Get the entry and skip it if it's in the exluded list */
		RegEntry = CONTAINING_RECORD(NextEntry, SMP_REGISTRY_VALUE, Entry);
		//wprintf(L"Processing known DLL: %ls-%ls\n", RegEntry->Name.Buffer, RegEntry->Value.Buffer);
		if ((SmpFindRegistryValue(&SmpExcludeKnownDllsList,
			RegEntry->Name.Buffer)) ||
			(SmpFindRegistryValue(&SmpExcludeKnownDllsList,
				RegEntry->Value.Buffer)))
		{
			continue;
		}

		/* Open the actual file */
		InitializeObjectAttributes(&ObjectAttributes,
			&RegEntry->Value,
			OBJ_CASE_INSENSITIVE,
			DirFileHandle,
			NULL);
		Status = NtOpenFile(&FileHandle,
			SYNCHRONIZE | FILE_EXECUTE,
			&ObjectAttributes,
			&IoStatusBlock,
			FILE_SHARE_READ | FILE_SHARE_DELETE,
			FILE_NON_DIRECTORY_FILE |
			FILE_SYNCHRONOUS_IO_NONALERT);
		if (!NT_SUCCESS(Status))
		{

			/* Go to the next entry */
			NextEntry = NextEntry->Flink;
			continue;
		}

		/* Checksum it */
		Status = LdrVerifyImageMatchesChecksum((HANDLE)((ULONG_PTR)FileHandle | 1),
			SmpProcessModuleImports,
			RegEntry,
			&ImageCharacteristics);
		if (!NT_SUCCESS(Status))
		{
			/* Checksum failed, so don't even try going further -- kill SMSS */
			RtlInitUnicodeString(&ErrorResponse,
				L"Verification of a KnownDLL failed.");
			ErrorParameters[0] = (ULONG)&ErrorResponse;
			ErrorParameters[1] = Status;
			ErrorParameters[2] = (ULONG)&RegEntry->Value;
			SmpTerminate(ErrorParameters, 5, RTL_NUMBER_OF(ErrorParameters));
		}
		else
			if (!(ImageCharacteristics & IMAGE_FILE_DLL))
			{
				/* An invalid known DLL entry will also kill SMSS */
				RtlInitUnicodeString(&ErrorResponse,
					L"Non-DLL file included in KnownDLL list.");
				ErrorParameters[0] = (ULONG)&ErrorResponse;
				ErrorParameters[1] = STATUS_INVALID_IMPORT_OF_NON_DLL;
				ErrorParameters[2] = (ULONG)&RegEntry->Value;
				SmpTerminate(ErrorParameters, 5, RTL_NUMBER_OF(ErrorParameters));
			}

		/* Temporarily hack the SD to use a default DACL for this section */
		if (SmpLiberalSecurityDescriptor)
		{
			OldFlag = SmpLiberalSecurityDescriptor->Control;
			SmpLiberalSecurityDescriptor->Control |= SE_DACL_DEFAULTED;
		}

		/* Create the section for this known DLL */
		InitializeObjectAttributes(&ObjectAttributes,
			&RegEntry->Value,
			OBJ_PERMANENT,
			DirHandle,
			SmpLiberalSecurityDescriptor)
			Status = NtCreateSection(&SectionHandle,
				SECTION_ALL_ACCESS,
				&ObjectAttributes,
				0,
				PAGE_EXECUTE,
				SEC_IMAGE,
				FileHandle);

		/* Undo the hack */
		if (SmpLiberalSecurityDescriptor) SmpLiberalSecurityDescriptor->Control = OldFlag;

		/* Check if we created the section okay */
		if (NT_SUCCESS(Status))
		{
			/* We can close it now, since it's marked permanent */
			Status1 = NtClose(SectionHandle);
			ASSERT(NT_SUCCESS(Status1));
		}
		else
		{
			/* If we couldn't make it "known", that's fine and keep going */
			RtlStringCchPrintfW(pszDest, cchDest, L"SMSS: CreateSection for KnownDll %ls failed - Status == %lx\n",
				RegEntry->Value.Buffer, Status);
		}

		/* Close the file since we can move on to the next one */
		Status1 = NtClose(FileHandle);
		ASSERT(NT_SUCCESS(Status1));

		/* Go to the next entry */
		NextEntry = NextEntry->Flink;
	}

Quickie:
	/* Close both handles and free the NT path buffer */
	if (DirHandle)
	{
		Status1 = NtClose(DirHandle);
		ASSERT(NT_SUCCESS(Status1));
	}
	if (DirFileHandle)
	{
		Status1 = NtClose(DirFileHandle);
		ASSERT(NT_SUCCESS(Status1));
	}
	if (NtPath.Buffer) RtlFreeHeap(RtlGetProcessHeap(), 0, NtPath.Buffer);
	return Status;
}


NTSTATUS
NTAPI
InitializeKnownDlls32(VOID)
{
	NTSTATUS Status;
	PSMP_REGISTRY_VALUE RegEntry;
	UNICODE_STRING DestinationString;
	PLIST_ENTRY Head, NextEntry;

	/* Call the internal function */
	RtlInitUnicodeString(&DestinationString, L"\\KnownDlls32");
	RtlInitUnicodeString(&SmpKnownDllPath, L"X:\\Windows\\SysWow64");
	Status = SmpInitializeKnownDllsInternal(&DestinationString, &SmpKnownDllPath);

	/* Wipe out the list regardless of success */
	Head = &SmpKnownDllsList;
	while (!IsListEmpty(Head))
	{
		/* Remove this entry */
		NextEntry = RemoveHeadList(Head);

		/* Free it */
		RegEntry = CONTAINING_RECORD(NextEntry, SMP_REGISTRY_VALUE, Entry);
		RtlFreeHeap(RtlGetProcessHeap(), 0, RegEntry->AnsiValue);
		RtlFreeHeap(RtlGetProcessHeap(), 0, RegEntry->Value.Buffer);
		RtlFreeHeap(RtlGetProcessHeap(), 0, RegEntry);
	}

	/* All done */
	return Status;
}



int main()
{
	NTSTATUS Status;
	PVOID OriginalEnvironment;
	ULONG MuSessionId = 0;
	/* Create the initial security descriptors */

	Status = SmpCreateSecurityDescriptors(TRUE);
	if (!NT_SUCCESS(Status))
	{
		/* Fail */
		SMSS_CHECKPOINT(SmpCreateSecurityDescriptors, Status);
		//return Status;
	}
	
	InitializeListHead(&SmpKnownDllsList);
	InitializeListHead(&SmpExcludeKnownDllsList);
	/* Initialize the SMSS environment */

	Status = RtlCreateEnvironment(TRUE, &SmpDefaultEnvironment);
	if (!NT_SUCCESS(Status))
	{
		/* Fail if there was a problem */
		RtlStringCchPrintfW(pszDest, cchDest, L"SMSS: Unable to allocate default environment - Status == %X\n",
			Status);
		SMSS_CHECKPOINT(RtlCreateEnvironment, Status);
		return Status;
	}

	/* Switch environments, then query the registry for all needed settings */
	OriginalEnvironment = NtCurrentPeb()->ProcessParameters->Environment;
	NtCurrentPeb()->ProcessParameters->Environment = SmpDefaultEnvironment;
	Status = RtlQueryRegistryValues(RTL_REGISTRY_CONTROL,
		L"Session Manager",
		SmpRegistryConfigurationTable,
		NULL,
		NULL);
	SmpDefaultEnvironment = NtCurrentPeb()->ProcessParameters->Environment;
	NtCurrentPeb()->ProcessParameters->Environment = (PWSTR)OriginalEnvironment;
	if (!NT_SUCCESS(Status))
	{
		/* We failed somewhere in registry initialization, which is bad... */
		RtlStringCchPrintfW(pszDest, cchDest, L"SMSS: RtlQueryRegistryValues failed - Status == %lx\n", Status);
		SMSS_CHECKPOINT(RtlQueryRegistryValues, Status);
		return Status;
	}
	Status = InitializeKnownDlls32();
	if (!NT_SUCCESS(Status))
	{
		return Status;
	}

	return STATUS_SUCCESS;
}

